from typing import Optional

import typer
from typing_extensions import Annotated


force_flag = Annotated[
    bool,
    typer.Option(
        "--force",
        "-f",
        help="Overwrite target files even if they were created manually.",
    ),
]
no_comment_flag = Annotated[
    bool,
    typer.Option(
        "--no-comment",
        "-C",
        help="Do not add the 'generated by yamlex' comment at the top of the file.",
    ),
]
verbose_flag = Annotated[
    bool,
    typer.Option(
        "--verbose",
        help="Enable verbose output.",
    ),
]
quiet_flag = Annotated[
    bool,
    typer.Option(
        "--quiet",
        help="Disable any informational output. Only errors.",
    ),
]
debug_flag = Annotated[
    bool,
    typer.Option(
        "--debug",
        hidden=True,
    ),
]
